/*
 * LICENSE:
 * Copyright (c) 2000-2001
 *  Charles D. Mercadal
 *   mercadal@diablonet.net
 *  All rights reserved.
 *	
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 3. Neither name of the Charles Mercadal nor the names of its contributors may be
 * used to endorse or promote products derived from this software without
 * specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY CHARLES MERCADAL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CHARLES MERCADAL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @(#) checks.c - 0.4 - Charles Mercadal - 10 December 2001
 */

#include "checks.h"

/* Function prototypes */
void	displayHelp(void);
void	initPrefs(pref *);
void	printDirectories(const info *, int);
void	doEventLoop(const pref *, info *);
void	parseChecksrc(info *);
void	flushLastVal(const info *, int);
int		getFileLength(void);
void	prefFilePath(char **);
char	*cmfgetln(FILE *);
char	*cmfgettoch(FILE *, const char);

int main(int argc, char *argv[])
{
    pref mainPref;			/* The main pref struct */
    info *dirinfo;			/* Array of directory info */
    register int counter;
    char usingp = 0;		/* Using the pref file .checksrc */
    int ch;
    extern char	*optarg;
    extern int optind;
    
    mainPref.minOrSec = 's';	/* Default value */
    mainPref.sleepTime = 60;	/* If none is given */
    mainPref.nrm = 0;			/* Alert when files removed */
    
    while ((ch = getopt(argc, argv, "m:s:pn")) != -1) {
        switch (ch) {
            case 'm':
                mainPref.minOrSec = 'm';
                mainPref.sleepTime = atoi(optarg);
                break;
            case 's':
                mainPref.minOrSec = 's';
                mainPref.sleepTime = atoi(optarg);
                break;
            case 'p':
                usingp = 1;
                break;
            case 'n':
                mainPref.nrm = 1;
                break;
            case '?':
            default:
                displayHelp();
                /* NOTREACHED */
                break;
        }
    }
    argc -= optind;
    argv += optind;
    
    if (usingp) {
        mainPref.numberOfDirs = getFileLength();
        
        if ((dirinfo = (info *)calloc(mainPref.numberOfDirs, sizeof(info))) == NULL)
            err(eMallocFoobarExit, BadMalloc);
            
        parseChecksrc(dirinfo);
    } else {
        if (argc <= 0)		/* no directory given */
            displayHelp();
        
        if ((dirinfo = (info *)calloc((mainPref.numberOfDirs = argc), sizeof(info))) == NULL )
            err(eMallocFoobarExit, BadMalloc);
            
        for (counter = 0; counter < argc; counter++) {
            dirinfo[counter].presentValue = -1;
            
            if ((dirinfo[counter].directory = strdup(argv[counter])) == NULL )
                err(eMallocFoobarExit, BadMalloc);
        }
    }
    
    /* Run printDirectories after initPrefs */
    initPrefs(&mainPref);
    printDirectories(dirinfo, mainPref.numberOfDirs);
    
    for (;;) {	
        for ((counter = mainPref.numberOfDirs - 1); counter >= 0; counter--)
            doEventLoop(&mainPref, &dirinfo[counter]);

        if (usingp)
            flushLastVal(dirinfo, mainPref.numberOfDirs);

        (void)sleep(mainPref.sleepTime);
    }
    
    /* NOTREACHED (ever) */
    return 1;
}
 
    
/*
 * displayHelp will inform uninformed users how 
 * to properly use the arguments available to checks.
 */
void displayHelp(void)
{

	(void)fprintf(stderr, "Usage:\tchecks [-n] [-ms delay time] [[-p] | [directories...]]\n-m:\tRequires delay in minutes\n-s:\tRequires delay in seconds\n-n:\tDo not alert if files are removed from directory\n-p:\tUse ~/.checksrc\nComments?  charleschuck@mac.com\n");
	exit(eBadUserExit);
}


/*
 * initPrefs will set up the pref struct with the values that
 * the other functions in the program expect, as well as tell the 
 * user that we are really actually checking the given directories
 * for them now.
 */
void initPrefs(pref *toAdjust)
{
    if (toAdjust->minOrSec == 'm') {
        if (toAdjust->sleepTime < 1)	/* if luser enters < 1 */
            toAdjust->sleepTime = 1;
        
        (void)printf("\nWill check the following folders approximately every %d minutes:\n", toAdjust->sleepTime);
        toAdjust->sleepTime = toAdjust->sleepTime * 60;
        toAdjust->minOrSec = 's';	/* sleep() takes values in seconds */
    } else {
        if ((toAdjust->sleepTime < 10) && (getuid() != 0))
            toAdjust->sleepTime = 10;	/* Don't let lusers grind disk too hard */
            
        (void)printf("\nWill check the following folders approximately every %d seconds:\n", toAdjust->sleepTime);
    }
}

/*
 * printDirectories just prints out the directories that 
 * are actually being checked.  Run after initPrefs.
 */
void printDirectories(const info *dirArray, register int counter)
{

    while (counter-- > 0)
        (void)puts(dirArray[counter].directory);
}


/*
 * doEventLoop is a poorly named function, as we really 
 * aren't getting any events, and the loop is external.  However, 
 * this function actually does the job of checking if the 
 * size of our directory has changed, and will print out
 * the message that it has (if it has).
 */
void doEventLoop(const pref *ourPrefs, info *dirinfo)
{
    off_t previous = dirinfo->presentValue;
    struct stat fileinfo;
    
    if (stat(dirinfo->directory, &fileinfo) != -1) {
    	dirinfo->presentValue = fileinfo.st_size;
    	
        if ((((previous > dirinfo->presentValue) && !(ourPrefs->nrm)) || previous < dirinfo->presentValue) && (previous != -1)) {
            (void)printf("\nFolder %s has changed\a\n", dirinfo->directory);
		}
	} else {
		perror(dirinfo->directory);
	}
}


/*
 * parseChecksrc will do it's darndest to read the .checksrc in
 * the user's home folder, extract directory names, and, if available, 
 * extract the size of the directory at last run through.
 */
void parseChecksrc(info *dinfo)
{
    FILE *infile;					/* parse this thing */
    char *stringloc;				/* pointer to read char * */
    register int counter;			/* general counter var */
    register int masterCounter = 0;	/* main counter for input */
    char success;
    char *prefLoc;

    prefFilePath(&prefLoc);
    
    if ((infile = fopen(prefLoc, "r")) == NULL)
        err(eBadFile, BadFileIO);
    
    free(prefLoc);

    while ((stringloc = cmfgetln(infile)) != NULL) {
		/* first character is white space or comment character */
        if (isspace(*stringloc) || (*stringloc == '#'))
            continue;

        for (counter = 0; stringloc[counter] != '\0'; counter++) {
            if (stringloc[counter] == '\t') {
                success = 1;
                
                if ((dinfo[masterCounter].directory = (char *)malloc(counter)) == NULL)
                    err(eMallocFoobarExit, BadMalloc);

				/* Copy string stringloc up to tabstop */
                strncpy(dinfo[masterCounter].directory, stringloc, counter);
                dinfo[masterCounter].directory[counter] = '\0';
                
                dinfo[masterCounter].presentValue = atoi(&stringloc[counter + 1]);
            }
        }

        if (!success) {	/* new directory entry */
            if ((dinfo[masterCounter].directory = strdup(stringloc)) == NULL)
                err(eMallocFoobarExit, BadMalloc);
            
            dinfo[masterCounter].presentValue = -1;
        }

        /* Get ready for next loop */
        free(stringloc);
        success = 0;
        ++masterCounter;
    }

    fclose(infile);
}


/*
 * flushLastVal will dump the values from the given info struct
 * into a checks readable file at .checksrc
 */
void flushLastVal(const info *toFlush, int howManyDirectories)
{
    sigset_t allSignals;
    FILE *outfile;
    char *prefLoc;	/* preference file location */

    prefFilePath(&prefLoc);
    sigfillset(&allSignals);	/* allSignals initialized with all signals */

    if ((outfile = fopen(prefLoc, "w")) == NULL)
        err(eBadFile, BadFileIO);
    
    free(prefLoc);
    
    /*
     * Careful!  We stop receiving proper signals here
     */
    sigprocmask(SIG_BLOCK, &allSignals, NULL);
    while (howManyDirectories > 0) {
        --howManyDirectories;
        
        (void)fprintf(outfile,
            "%s\t%ld\n",
            toFlush[howManyDirectories].directory,
            (long)toFlush[howManyDirectories].presentValue);
    }
    
    fclose(outfile);
    sigprocmask(SIG_UNBLOCK, &allSignals, NULL);
    /*
     * ...and here's where proper signals return
     */
}


/*
 * This function will return the length of a file, well, really
 * how many parsable lines there are in ~/.checksrc
 */
int getFileLength(void)
{
    register int counter = 0;
    FILE *infile;
    char *firstChar;	/* char * from fgetln */
    char *prefLoc;

    prefFilePath(&prefLoc);
        
    if ((infile = fopen(prefLoc, "r")) == NULL)	/* Damn...no .checksrc */
        err(eBadFile, BadFileIO);
        
    free(prefLoc);	/* after use, free what prefFilePath returns */

    while ((firstChar = cmfgetln(infile))) {
        if (isspace(*firstChar) || (*firstChar == '#'))
            continue;
                
        ++counter;
    }
    fclose(infile);

    return counter;
}


/*
 * prefFilePath gets the path of .checksrc and sets toSet to that value.
 */
void prefFilePath(char **toSet)
{
    if ((*toSet = (char *)malloc(strlen(getenv("HOME")) + strlen(PrefFile) + 1)) == NULL)
        err(eMallocFoobarExit, BadMalloc);

    (void)sprintf(*toSet, "%s%s", getenv("HOME"), PrefFile);
}


/*
 * cmfgetln is the easy-to-use version of cmfgettoch for 
 * getting a new line quick.  See notes before cmfgettoch.
 */
char *cmfgetln(FILE *ifp)
{

    return cmfgettoch(ifp, '\n');
}


/*
 * cmfgettoch will read from the FILE pointer ifp until the character to is found, and will 
 * return a string of all the data thereto, replacing the character to with a \0 character.  
 * Will return NULL if called during an EOF condition.
 * 
 * Buffers allocated here should be freed by whatever function uses it.
 */
char *cmfgettoch(FILE *ifp, const char to)
{
    #ifdef BSD	/* create BSD optimized cmfgettoch */
    register unsigned char *pos;
    char *ret;
    extern int __srefill __P((FILE *));
    register size_t length;
    

    /*
     * Do we have any more input?  If we haven't,
     * we should probably return some sort of value 
     * stating that.
     */
    if (ifp->_r <= 0 && __srefill(ifp))
        return NULL;
                
    /*
     * So, we search for our character.  If one is found, 
     * we figure the length of the string in relation
     * to fp->_p.  If not, we figure it in relation to 
     * fp->_r (what's left in the file).
     * If you're confused, /usr/include/stdio.h has the 
     * FILE type definition on Berkeley systems.
     */
    if ((pos = memchr((void *)ifp->_p, to, ifp->_r )))
        length = (++pos - ifp->_p);
    else
        length = (ifp->_r + 1);

    if((ret = (char *)malloc(length * sizeof(char))) == NULL) {
        perror(NULL);	/* bail out, no more VM */
        exit(1);
    }
    ret = memcpy(ret, (void *)ifp->_p, length);
    ret[length - 1] = '\0';
    ifp->_p = pos;
    ifp->_r -= length;
    return ret;
    
    #else
    register int ch;
    register char *buf;
    register size_t blen, cnt;
    
    
    cnt = 0;
    blen = 40;
    buf = (char *)malloc(blen);
    
    while (((ch = fgetc(ifp)) != to) || (ch == EOF)) {
        buf[cnt] = ch;
        
        /* whoa...end of file.  Time to stop */
        if (ch == EOF)
            break;
            
        cnt++;
        
        /* out of buffer space...reallocate */
        if (cnt >= blen) {
            blen += 40;
            if ((buf = (char *)realloc(buf, blen)) == NULL) {
                    perror(NULL);
                    exit(1);
            }
        }
    }
    
    /* Return NULL terminated string */
    buf[cnt] = '\0';
    
    /* No more lines...return NULL pointer */
    if((cnt == 0) && (ch == EOF)) {
        free(buf);
        return NULL;
    }
    
    /* All is OK, return the string */
    return buf;
    #endif	/* BSD */
}
